unit Main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Menus, Grids,
  Base, PuzzleIO;

type
  TForm1 = class(TForm)
    BoxGrid: TDrawGrid;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    Exit1: TMenuItem;
    N1: TMenuItem;
    PrintSetup1: TMenuItem;
    Print1: TMenuItem;
    N2: TMenuItem;
    SaveAs1: TMenuItem;
    Save1: TMenuItem;
    Open1: TMenuItem;
    New1: TMenuItem;
    Label1: TLabel;
    Label2: TLabel;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    PiecesGrid: TStringGrid;
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject);
    procedure Open1Click(Sender: TObject);
    procedure BoxGridClick(Sender: TObject);
    procedure BoxGridDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure PiecesGridDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure PiecesGridMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  private
    { Private declarations }
  public
    { Public declarations }
    FPuzzle: TPuzzle;
    procedure DoReadPuzzle(AFileName: String);
    procedure ClearPiecesGrid;
    procedure FillPiecesGrid;
    procedure NextOrientationInPiecesGrid(APlacement: TPlacement);
    procedure UpdatePiecesGrid;
    procedure PlacePieceInBox(const APieceName: String;
      AOrientationIndex: Integer; ARowIndex, AColIndex: Integer);
    procedure UpdateBoxGrid;
    procedure UpdateViews;
  end;

const
  fnDefaultPuzzleFolder = '..\Puzzles';
  fnDefaultPuzzleFileName = 'SimplePuzzle.txt';

var
  Form1: TForm1;

implementation //===============================================================

uses
  Math;

{$R *.dfm}

{ TForm1 }

procedure TForm1.ClearPiecesGrid;
  // Free all placements on PiecesGrid
var
  ColIndex, RowIndex: Integer; // to traverse PiecesGrid.Objects
  VPlacement: TPlacement; // in PiecesGrid at Col, Row
  VPositionList: TPositionList; // to capture result of CoverP
  J: Integer; // to traverse VPositionList
begin
  with PiecesGrid do
  begin

    for RowIndex := 0 to RowCount - 1 do
      for ColIndex := 0 to ColCount - 1 do
      begin
        VPlacement := Objects[ColIndex, RowIndex] as TPlacement;
        if VPlacement <> nil then
        begin
          VPositionList := CoverP(VPlacement);

          for J := 0 to Length(VPositionList) - 1 do
            with VPositionList[J] do
            begin
              Form1.PiecesGrid.Objects[Col, Row] := nil;
            end; { with }

        end; { if }
        VPlacement.Free;
        Cells[Col, Row] := '';
      end; { for RowIndex }
    end; { with }
end;

procedure TForm1.FillPiecesGrid;
var
  VPos: TPosition; // current position in PiecesGrid
  VRowCount, VColCount: Integer; // to determine size of PiecesGrid
  I: Integer; // to traverse the pieces
  VPiece: TPiece; // piece being placed
  VOrientation: TOrientation; // first orientation of VPiece
  VPieceSize: Integer; // max width of VPiece orientation
  VPlacement: TPlacement; // of VPiece
  VPositionList: TPositionList; // to capture result of CoverP
  J: Integer; // to traverse VPositionList
begin
  VColCount := 1;
  VRowCount := 1;
  VPos.Col := 1;
  VPos.Row := 2; // will not change

  // calculate placements for pieces in PiecesGrid, depending on their max size
  with FPuzzle.Stock do
  begin
    for I := 0 to Count - 1 do
    begin
      VPiece := GetPiece(I);
      VOrientation := VPiece.GetOrientation(0); // initial orientation
      VPlacement := TPlacement.Create(VPiece, VOrientation, VPos);

      // determine size of square bounding box for all orientations
      VPieceSize := Max(VOrientation.MaxCol + 1, VOrientation.MaxRow + 1);
      VPos.Col := VPos.Col + VPieceSize + 1;
      // update dimensions of PiecesGrid
      VRowCount := Max(VRowCount, 2 + VPieceSize + 1);
      VColCount := VPos.Col;

      // set dimensions for Piecesgrid
      PiecesGrid.ColCount := VColCount;
      PiecesGrid.RowCount := VRowCount;

      // Assign placement to corresponding cells in PiecesGrid,
      VPositionList := CoverP(VPlacement);

      for J := 0 to Length(VPositionList) - 1 do
        with VPositionList[J] do
        begin
          PiecesGrid.Objects[Col, Row] := VPlacement;
        end; { with }

      // write name on top
      with VPositionList[0] do
        PiecesGrid.Cells[Col, 0] := VPlacement.FPiece.Name;
    end; { for I }

  end; { with }
end;

procedure TForm1.NextOrientationInPiecesGrid(APlacement: TPlacement);
var
  VPositionList: TPositionList; // positions of old/new orientation
  J: Integer; // to traverse VPositionList
  I: Integer; // to traverse orientations of APlacement.FPiece
  VIndex: Integer; // to determine index of next orientation
begin
  // remove references to old placement from PiecesGrid
  VPositionList := CoverP(APlacement);

  for J := 0 to Length(VPositionList) - 1 do
    with VPositionList[J] do
    begin
      PiecesGrid.Objects[Col, Row] := nil;
    end;{with}

  // determine new orientation as successor of old orientation
  // in orientation list of piece
  with APlacement, FPiece do
  begin
    VIndex := -1;

    for I := 0 to OrientationCount - 1 do
      if GetOrientation(I) = FOrientation
      then VIndex := (I + 1) mod OrientationCount;

    FOrientation := GetOrientation(VIndex);
  end; {with APlacement, FPiece}

  // put references to new placement in PiecesGrid
  VPositionList := CoverP(APlacement);

  for J := 0 to Length(VPositionList) - 1 do
    with VPositionList[J] do
    begin
      PiecesGrid.Objects[Col, Row] := APlacement;
    end;{with}
end;


procedure TForm1.DoReadPuzzle(AFileName: String);
begin
  ClearPiecesGrid;
  FPuzzle.Free;
  FPuzzle := ReadPuzzle(AFileName);
end;


procedure TForm1.UpdateBoxGrid;
begin
  BoxGrid.RowCount := FPuzzle.Box.RowCount;
  BoxGrid.ColCount := FPuzzle.Box.ColCount;
  BoxGrid.Refresh;
end;

procedure TForm1.UpdatePiecesGrid;
begin
  PiecesGrid.Refresh;
end;

procedure TForm1.UpdateViews;
begin
  UpdatePiecesGrid;
  UpdateBoxGrid;
end;

procedure TForm1.PlacePieceInBox(const APieceName: String;
  AOrientationIndex: Integer; ARowIndex, AColIndex: Integer);
var
  VPiece: TPiece; // piece to be placed in box, name is APieceName
  VOrientation: TOrientation; // orientation of VPiece for AOrientationIndex
  VPosition: TPosition; // position in box at ARowIndex, AColIndex
  VPlacement: TPlacement; // (VPiece, VOrientation, VPosition)
begin
  VPiece := FPuzzle.Stock.FindPiece(APieceName);
  VOrientation := VPiece.GetOrientation(AOrientationIndex);
  VPosition.Row := ARowIndex;
  VPosition.Col := AColIndex;
  VPlacement := TPlacement.Create(VPiece, VOrientation, VPosition);
  FPuzzle.DoPlacement(VPlacement);
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  ChDir(fnDefaultPuzzleFolder);
  DoReadPuzzle(fnDefaultPuzzleFileName);
  FillPiecesGrid;

// Temporarily, for testing purposes ------------------------------------------
// Assume: Simple Puzzle is loaded
    PlacePieceInBox('A', 0, 0, 1);
    PlacePieceInBox('B', 1, 0, 0);
//-----------------------------------------------------------------------------

  UpdateViews;
end;

procedure TForm1.FormClose(Sender: TObject);
begin
  Close;
end;

procedure TForm1.Open1Click(Sender: TObject);
begin
  if OpenDialog1.Execute then
  begin
    DoReadPuzzle(OpenDialog1.FileName);
    FillPiecesGrid;
    UpdateViews;
  end; { if }
end;

procedure TForm1.BoxGridClick(Sender: TObject);
begin
  UpdateViews;
  with BoxGrid do
  begin
//    Canvas.Brush.Color := clRed;
//    Canvas.FillRect(CellRect(Col,Row));
  end;
end;

procedure TForm1.BoxGridDrawCell(Sender: TObject; ACol, ARow: Integer;
            Rect: TRect; State: TGridDrawState);
var
  VColor: TColor; // color for grid cell
  VState: TCellState; // cell state at ARow, ACol
begin
  with FPuzzle.Box do
  begin
    if ValidCoordinates(ARow, ACol) then
    begin
      VState := FPuzzle.Box.CellState(ARow, ACol);
      case VState of
        csFree:
          begin
            VColor := clWhite;
          end;
        csOccupied:
          begin
            VColor := FPuzzle.Box.CellOccupant(ARow, ACol).FPiece.Color;
          end;
        csBlocked:
          begin
            VColor := clBlack;
          end;
      end;{case}
     end;{if}
    end;{with}

  BoxGrid.Canvas.Brush.Color := VColor;
  BoxGrid.Canvas.FillRect(Rect);
end;

procedure TForm1.PiecesGridDrawCell(Sender: TObject; ACol, ARow: Integer;
            Rect: TRect; State: TGridDrawState);
var
  VPiece: TPiece; // piece at ARow, ACol
  VColor: TColor; // color of VPiece
begin
  with PiecesGrid do
    if Cells[ACol, ARow] = '' then
    // tricky, assumes that text cells and piece cells are disjoint
    begin
      // determine color
      if Objects[ACol, ARow] = nil
      then VColor := clWhite
      else
        begin
          VPiece := (Objects[ACol, ARow] as TPLacement).FPiece;
          with FPuzzle.Stock do
            if VPiece.Multiplicity > 0
            then VColor := VPiece.Color
            else VColor := clLtGray;
        end;

      // draw cell
      Canvas.Brush.Color := VColor;
      Canvas.FillRect(Rect);
    end;{with}
end;

procedure TForm1.PiecesGridMouseDown(Sender: TObject; Button: TMouseButton;
            Shift: TShiftState; X, Y: Integer);
var
  VCol, VRow: Integer; // to determine cell coordinates of mouse down
begin
  with Sender as TStringGrid do
  begin
    // determine which cell was clicked
    MouseToCell(X, Y, VCol, VRow);
    if Objects[VCol, VRow] <> nil then // clicked on a piece
    begin
      case Button of
        mbRight:
          begin
            // determine next orientation of piece and update PiecesGrid
            NextOrientationInPiecesGrid(Objects[VCol, Vrow] as TPlacement);
            UpdatePiecesGrid;
          end; { mbRight }
      else // other modifiers
        // ignore
      end; { case }
    end
    else // not clicked on a piece
      // ignore
  end; { with Sender }
end;

end.
